/* 
 * icv150.c 
 *
 * ICV150 - kernel module DEVICE DRIVER
 *	    for the ADAS ICV150 Analog Acquisition board
 * by Integrated Real Time Systems (IRTS) - julien23@multimania.com
 * for European Synchrotron Radiation Facility (ESRF)
 *
 * OS : VME Linux/m68k 2.0 and 2.2
 *
 * VERSION : 1.0
 *
 * This device driver makes an easy access to the analog input values of the
 * ADAS ICV150 Analog Acquisition board via IOCTL() calls.
 *
 * 08/11/1999 (0.2) : Added RECALL command in ioctl().
 *
 * 15/11/1999 (0.3) : Removed Mutex protection from driver. They were useless
 *			as the driver never call scheduler (sleep...)
 *
 * 23/11/1999 (0.4) : Moved ICV150_regs[] declaration
 *
 */



/*----------------------------------------------------------------------
 * Define constants
 *---------------------------------------------------------------------*/

#ifndef __KERNEL__		/*kernel optimization*/
#	define __KERNEL__
#endif

#ifndef MODULE			/*always a module*/
#	define MODULE
#endif



/*----------------------------------------------------------------------
 * Include files 
 *---------------------------------------------------------------------*/

#include <linux/config.h>	/*required for device drivers*/
#include <linux/module.h>	/*module optimization*/
#include <linux/version.h>	/*module needs to be in the good version*/ 
#include <linux/kernel.h>	/*kernel calls*/	
#include <linux/errno.h>	/*error flags*/
#include <linux/malloc.h>	/*pointers management*/
#include <linux/fs.h>		/*struct inode and struct file declaration*/
#include <linux/ioport.h>	/*check, request and release I/O region*/
#include <linux/wrapper.h>	/*sleeping functions*/

#include "icv150.h"
#include "probelib.h"



/*----------------------------------------------------------------------
 * Macros 
 *---------------------------------------------------------------------*/

/*
 * These macros are used to determine which kernel you are using and to
 * compil the driver with the appropriate kernel calls
 */
#ifndef VERSION_CODE
#  define VERSION_CODE(vers,rel,seq) ( ((vers)<<16) | ((rel)<<8) | (seq) )
#endif

#if LINUX_VERSION_CODE < VERSION_CODE(2,0,0)	/*kernel < 2.0 - doesn't work*/
#  error "This kernel is too old: not supported by this file"
#endif
#if LINUX_VERSION_CODE < VERSION_CODE(2,1,0)
#  define LINUX_2_0	/*kernel 2.0*/
#else			/*kernel 2.1*/
#  include <asm/uaccess.h>	/*copy_to/from_user routine for kernel>2.1*/
#endif



/*----------------------------------------------------------------------
 * Local variables
 *---------------------------------------------------------------------*/

static u8		ICV150_debug	= 5;	/*debug level (default->1)*/
static u8		ICV150_boards	= 0;	/*number of boards (default->0)
						 *this number is fixed during
						 *the extract of parameters*/
static u16		ICV150_major	= 0;	/*0 (default) -> dynamique major
						 *allocation*/
static ICV150_STATE	*ICV150_state;	/*this struct is used to save important
					 *parameters about each board*/

static ICV150_REGS	*ICV150_regs[MAX_BOARDS];


/*----------------------------------------------------------------------
 * Function: ICV150_init()
 *
 * Makes the ICV150 Device Driver's initialization. It is called by the
 * init_module routine in order to check that hardware is well configured
 * and to reserve memory and IRQ vectors for the driver.
 *
 * OUTPUT :	0 -> init has succeed
 *		>0 -> init has failed (see asm/errno.h for value's means)
 *---------------------------------------------------------------------*/

static int ICV150_init()
{
	int		err, i, k;

/*
 * Check if the I/O regions you request are not already used by a device
 */
	for (i=0; i<ICV150_boards; i++)
	{
		err = check_region((u32)io_adr[i], ICV150_SIZE);
		if ( err )
		{
			if (ICV150_debug > 0)
				printk("  -The IO address space you indicate is already used\n");
			goto region_error; /*may be some region has already been
					*requested and 'i' is the number of those*/
		}
		request_region((u32)io_adr[i], ICV150_SIZE, "ICV150");
	}

/*
 * Allocate the memory for ICV150_state stucture
 */
	ICV150_state = (ICV150_STATE *) kmalloc(sizeof(ICV150_STATE) * ICV150_boards, GFP_KERNEL);
	if ( ICV150_state == NULL )
	{
		if (ICV150_debug > 0)
			printk("  -Unable to get enough memory to drive the ICV150 device");
		err = -ENOMEM;
		i = ICV150_boards;	/*all the regions have been requested*/
		goto region_error;	/*region has already been requested!*/
	}

	/*set all the ICV150_state structure to 0*/
	memset(ICV150_state, 0, sizeof(ICV150_STATE) * ICV150_boards);

/*
 * Reset the board and set the init values of structs
 */
	for (i=0; i<ICV150_boards; i++)
	{
		ICV150_regs[i] = (ICV150_REGS *) io_adr[i];  /*map the board*/
		ICV150_regs[i]->Stop_Status = 0; /*read or write this register
						  *stop scanning*/
		ICV150_regs[i]->RWit = 0; /*initialize the interrupt register*/
		ICV150_regs[i]->Channels = DEF_CHANNELS; /*initialize number*/
		(ICV150_state+i)->Channels = DEF_CHANNELS; /*save it*/
		ICV150_regs[i]->Start = 0;	/*read or write this register
					 	start automatic scanning*/
	}

/*
 * If everything have succeed we can continue
 */
	return 0;

/*
 * Something has failed
 */
	region_error:	/*'i' contains the board number where request has failed
			 *the former ones have already been requested*/
		for (k=0; k<i; k++)
			release_region((u32)io_adr[k], ICV150_SIZE);

		return err;
}



/*----------------------------------------------------------------------
 * Function: ICV150_ioctl()
 *
 * Handle the ioctl() calls. This routine realizes the most part of the
 * Device Driver's work. It is the only way to make an access to the
 * ICV150 I/O board.
 *
 * INPUT:	FD  ->	File descriptor
 *		CMD ->	RD_CHANNEL
 *			RD_ALL_CHANNELS 
 *			RD_GAIN
 *			RD_ALL_GAINS
 *			SET_NUMBER
 *			SET_GAIN
 *			SET_ALL_GAINS
 *			START
 *			EXT_TRIG
 *			STOP
 *			STORE
 *			RECALL
 *			DEBUG
 *		ARG ->	Array of MAX_CHANNELS u16
 *
 * OUTPUT :	0 -> command has succeed
 *		>0 -> command has failed (see asm/errno.h for value's means)
 *---------------------------------------------------------------------*/

static int ICV150_ioctl (inode, file, cmd, arg)
struct inode *inode;
struct file *file;
int cmd;
u16 arg[MAX_CHANNELS]; 
{
	u8		minor, num, i;
	u16		_arg[MAX_CHANNELS], out=0;
	ICV150_STATE	*state;
	ICV150_REGS	*regs;

/*
 * Determine which board is going to be accessed and then initialize
 * the local pointers that give us access to the board status and registers.
 * The minor number indicates which board is concerned : 0 for the first board
 * ... 3 for the last one (fourth)
 */
	minor = MINOR(inode->i_rdev);	/*get minor*/

	if ( minor >= ICV150_boards )	/*is it a correct minor?*/
	{
		if (ICV150_debug > 0)
			printk("  -ICV150_ioctl() error : Bad minor number");
		return -ENODEV;
	}

	state = ICV150_state + minor;	/*local pointer to status struct*/
	regs = ICV150_regs[minor];	/*local pointer to board's registers struct*/

	if (ICV150_debug > 1)
		printk("  -ICV150_ioctl() : board %d\n", minor);
	if (ICV150_debug > 2)
		printk("\t.cmd = %#X\n\t.board status = %#04X\n\t.ext status = %#04X\n", cmd, regs->Stop_Status, regs->Ext);

/*
 * Copy argument from user memory space to kernel memory space. We have to
 * do that to get user arguments. The routine has been optimized on kernel
 * 2.1 and higher
 */
	if ( (cmd == RD_CHANNEL) || (cmd == RD_GAIN) || (cmd == SET_NUMBER)
		|| (cmd == SET_GAIN) || (cmd == SET_ALL_GAINS) || (cmd == DEBUG) )
	{
#ifdef LINUX_2_0
		if (verify_area(VERIFY_READ, arg, MAX_CHANNELS*sizeof(u16)) == -EFAULT)
			return -EFAULT;
		memcpy_fromfs (_arg, arg, MAX_CHANNELS*sizeof(u16));
#else
		if ( copy_from_user(_arg, arg, MAX_CHANNELS*sizeof(u16)) )
			return -EFAULT;
#endif
	}

	switch (cmd)
	{
	case RD_CHANNEL:
	/*
	 * Check if the channel number is valid
	 */
		if ( _arg[0] >= state->Channels )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : channel out of range\n");
			return -EINVAL;
		}
		num = _arg[0];

	/*
	 * Check if the board is scanning or waiting for external trig
	 */
		if ( ((regs->Stop_Status & 1) == 0) && (state->External == 0) )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : board is not scanning\n");
			return -EAGAIN;
		}

		_arg[0] = regs->Data[num];	/*read the value*/
		
		out = 1;
		break;

	case RD_ALL_CHANNELS:
	/*
	 * Check if the board is scanning or waiting for external trig
	 */
		if ( ((regs->Stop_Status & 1) == 0) && (state->External == 0) )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : board is not scanning\n");
			return -EAGAIN;
		}

		for (i=0; i<state->Channels; i++)
			_arg[i]  = regs->Data[i];	/*read the values*/

		out = state->Channels;
		break;

	case RD_GAIN:
	/*
	 * Check if the channel number is valid
	 */
		if ( _arg[0] >= state->Channels )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : channel out of range\n");
			return -EINVAL;
		}
		num = _arg[0];

	/*
	 * Check if the board is not scanning or waiting for external trig
	 */
		if ( ((regs->Stop_Status & 1) == 1) || (state->External == 1) )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : board is scanning\n");
			return -EBUSY;
		}

		_arg[0] = (regs->Gain[num] & 0xF);	/*read the gain value*/

		out = 1;
		break;

	case RD_ALL_GAINS:
	/*
	 * Check if the board is not scanning or waiting for external trig
	 */
		if ( ((regs->Stop_Status & 1) == 1) || (state->External == 1) )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : board is scanning\n");
			return -EBUSY;
		}

		for (i=0; i<state->Channels; i++)
			_arg[i] = (regs->Gain[i] & 0xF); /*read the gain values*/

		out = state->Channels;
		break;

	case SET_NUMBER:
	/*
	 * Check if the number of channels is valid
	 */
		if ( (_arg[0] <= 0) && (_arg[0] >= MAX_CHANNELS) ) 
		{
			if (ICV150_debug > 2)
				printk("\t.Error : channel out of range\n");
			return -EINVAL;
		}

	/*
	 * Check if the board is not scanning or waiting for external trig
	 */
		if ( ((regs->Stop_Status & 1) == 1) || (state->External == 1) )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : board is scanning\n");
			return -EBUSY;
		}

	/*
	 * Set the number of channels to scan and save the number
	 */
		regs->Channels =  _arg[0]-1;	/*0 for one channel, 1 for two
						 *channels...*/
		state->Channels = _arg[0];	/*real number*/

		return 0;

	case SET_GAIN:
	/*
	 * Check if the channel number is valid
	 */
		if ( _arg[0] >= MAX_CHANNELS )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : channel out of range\n");
			return -EINVAL;
		}

	/*
	 * Check if the gain is in a correct range
	 */
		if ( _arg[1] > MAX_GAIN )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : gain out of range\n");
			return -EINVAL;
		}

	/*
	 * Check if the board is not scanning or waiting for external trig
	 */
		if ( ((regs->Stop_Status & 1) == 1) || (state->External == 1) )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : board is scanning\n");
			return -EBUSY;
		}

		regs->Gain[_arg[0]] = _arg[1];	/*set gain*/

		return 0;

	case SET_ALL_GAINS:
	/*
	 * Check if the board is not scanning or waiting for external trig
	 */
		if ( ((regs->Stop_Status & 1) == 1) || (state->External == 1) )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : board is scanning\n");
			return -EBUSY;
		}

		for (i=0; i<state->Channels; i++)
		{
			if ( _arg[i] > MAX_GAIN )
			{
				if (ICV150_debug > 2)
					printk("\t.Error : gain out of range\n");
				return -EINVAL;	
			}

			regs->Gain[i] = _arg[i];	/*set gain*/
		}

		return 0;

	case START:
		regs->Stop_Status = 0;	/*stop scanning*/
		regs->RWit = 0;		/*disable interrupts*/
		regs->Start = 0;	/*start continuous scanning*/
		state->External = 0;	/*save the status*/
		return 0;

	case EXT_TRIG:
		regs->Stop_Status = 0;	/*stop scanning*/
		regs->RWit = 0;		/*disable interrupts*/
		regs->Ext = 0;		/*start scanning on external trig*/
		state->External = 1;	/*save the status*/
		return 0;

	case STOP:
		regs->Stop_Status = 0;	/*stop scanning (all sorts of scan)*/
		state->External = 0;	/*clear external*/
		return 0;

	case STORE:
	/*
	 * Check if the board is not scanning or waiting for external trig
	 */
		if ( ((regs->Stop_Status & 1) == 1) || (state->External == 1) )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : board is scanning\n");
			return -EBUSY;
		}

		regs->Store = 0;	/*store all the gains in the EEPROM*/

		return 0;

	case RECALL:
	/*
	 * Check if the board is not scanning or waiting for external trig
	 */
		if ( ((regs->Stop_Status & 1) == 1) || (state->External == 1) )
		{
			if (ICV150_debug > 2)
				printk("\t.Error : board is scanning\n");
			return -EBUSY;
		}

		regs->Array = 0;	/*retrieve all the gains in the EEPROM*/

		return 0;

	case DEBUG:
		ICV150_debug = _arg[0];
		return 0;

	default:
		if (ICV150_debug > 1)
			printk("\t.Bad ioctl() command\n");
		return -EINVAL;
	}

/*
 * Copy back the arg structure that contain the result of the read operation
 */
	if ( out )
	{
#ifdef LINUX_2_0
		if (verify_area(VERIFY_WRITE, arg, out*sizeof(u16)) == -EFAULT)
			return -EFAULT;
		memcpy_tofs (arg, _arg, out*sizeof(u16));
#else
		if ( copy_to_user(arg, _arg, out*sizeof(u16)) )
			return -EFAULT;
#endif
	}
		return 0;
}



/*----------------------------------------------------------------------
 * Function: ICV150_open()
 *
 * Called when /dev/ICV150_x is opened. Update the usage count so the module
 * doesn't get unloaded while it is in use. 
 *---------------------------------------------------------------------*/

static int ICV150_open (inode, file)
struct inode *inode;
struct file *file;
{
	if (ICV150_debug > 1)
		printk("  -ICV150_open() : PID = %d\n", current->pid);

	MOD_INC_USE_COUNT;	/*one more process/thread plays with us*/
	return 0;
}



/*----------------------------------------------------------------------
 * Function: ICV150_release()
 *
 * Called when /dev/ICV150_x is closed or when the process exits. Update the
 * usage count so that the module get unloaded if no more process need it
 * in a certain amount of time. In kernel 2.1 or higher this function
 * return a value (alowed to fail now?)
 *---------------------------------------------------------------------*/
#ifdef LINUX_2_0
static void ICV150_release(inode, file)
#else
static int ICV150_release(inode, file)
#endif
struct inode *inode;
struct file *file;
{
	if (ICV150_debug > 1)
		printk("  -ICV150_release() : PID = %d\n", current->pid);

	MOD_DEC_USE_COUNT;	 /*one more process/thread leaves the game*/

#ifdef LINUX_2_0
	/*nothing to return*/
#else
	return 0;
#endif
}



/*----------------------------------------------------------------------
 * Device driver entries points 
 *---------------------------------------------------------------------*/

struct file_operations ICV150_fops =
{
	NULL,			/* _seek	*/
	NULL,			/* _read	*/
	NULL,			/* _write	*/
	NULL,			/* _readdir	*/
	NULL,			/* _select	*/
	ICV150_ioctl,		/* _ioctl	*/
	NULL,			/* _mmap	*/
	ICV150_open,		/* _open	*/
	ICV150_release,		/* _release	*/
};



/*----------------------------------------------------------------------
 * Function: init_module()
 *
 * Initialize the module
 *---------------------------------------------------------------------*/

int init_module()
{
	int err, err1, i=0;

	if (ICV150_debug > 0)
		printk(KERN_INFO "ICV150 Device Driver\n");

/*
 * Extract and check loading parameters.
 */
	while ( (i < MAX_BOARDS) && (io_adr[i] != NULL) )
	{
		io_adr[i] = (u32 *)((u32)io_adr[i] << 8);

	/*
	 * Probe the ICV150 board by reading at the first and the last addresses
	 * of the board
	 */
		err = probe((char*)io_adr[i], sizeof(short), -1);
		err1 = probe(((char*)io_adr[i])+ICV150_SIZE-2, sizeof(short), -1);
		if (err || err1)
		{
			if (ICV150_debug > 0)
				printk("Board %d is not at the specified address\n", i);
			return -EFAULT;
		}

	/*
	 * Set default base vector if not specified
		if ( int_vector[i] == 0 )
			int_vector[i] = int_vector[i-1] + 1;
  	 */
		ICV150_boards++;
		i++;
	}

/*
	if ( (int_level <= 0) || (int_level > 7) )
	{
		printk("  -Bad VME IRQ level\n");
		return -EINVAL;
	}
*/

/*
 * Register the ICV150 (char) Device Driver with the ICV150_major number. If
 * ICV150_major is set to 0 then the allocation number will be determined by
 * the kernel
 */
	err = register_chrdev(ICV150_major, "ICV150", &ICV150_fops);
	if ( err < 0 )	/*register_chrdev has failed*/
	{
		if (ICV150_debug > 0)
			printk("  -Unable to get major for ICV150 device");
		return err;
	}
	if ( ICV150_major == 0)		/*dynamic major allocation*/
		ICV150_major = err;

/*
 * If error has appeared during ICV150_init, we quit with error
 */
	err = ICV150_init();
	if ( err )
	{
		unregister_chrdev(ICV150_major, "ICV150");
		return err;
	}

	if (ICV150_debug > 1)
		printk("  -%d boards have been successfully initialized\n", ICV150_boards);

	return 0;
}



/*----------------------------------------------------------------------
 * Function: cleanup_module()
 *
 * Undid whatever init_module did
 *---------------------------------------------------------------------*/

void cleanup_module()
{
	int i;

/*
 * Free all irq vectors
	for (i=0; i<ICV150_boards; i++)
	{
		ICV150_STATE *state = (ICV150_state + i);

		free_irq(int_vector[i], state);
	}
 */

/*
 * Release the allocated memory
 */
	kfree(ICV150_state);

/*
 * Release the I/O board address region
 */
	for (i=0; i<ICV150_boards; i++)
		release_region((u32)io_adr[i], ICV150_SIZE);

 /*
  * Unregister the device before leaving
  */
	unregister_chrdev(ICV150_major, "ICV150");

	if (ICV150_debug > 0)
		printk(KERN_INFO "End of ICV150 Device Driver\n\n");
}
